// -*- mode: c++; indent-tabs-mode: nil -*-
//! @file YamlRpcHandler.qm @brief YAML-RPC handler class definition for the HttpServer module

/*  YamlRpcHandler.qm Copyright (C) 2012 - 2015 David Nichols

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

// minimum qore version

// requires the binary yaml module

// need mime definitions

// need Util module

// need HttpServerUtil module at a minimum in this module

// do not use $ signs in declarations


/*  Version History
    * 2014-03-05 v1.2: David Nichols <david@qore.org>: minor error logging updates, updated minimum qore version to 0.8.7 for get_ex_pos()
    * 2013-09-24 v1.1: David Nichols <david@qore.org>: added optional log closure to constructor
    * 2012-05-24 v1.0: David Nichols <david@qore.org>: updated to a user module
*/

/** @mainpage YamlRpcHandler Module

    @section yamlrpchandlerintro Introduction

    This module implements server-side support for a proprietary RPC-based protocol using <a href="http://yaml.org">YAML</a> for serialization/deserialization of message data.

    This module provides the @ref YamlRpcHandler::YamlRpcHandler "YamlRpcHandler" class which can be used to provide an RPC handler for the HttpServer class provided by the HttpServer module.

    See the <a href="../../YamlRpcClient/html/index.html">YamlRpcClient module</a> for a description of the YAML-RPC protocol implemented here.

    @section example Example Usage
    @code
%requires HttpServer
%requires YamlRpcHandler
%requires Mime

const ApiMethods = (
     ("name": "^sys\\.shutdown\$",
      "text": "sys.shutdown",
      "function": sub () { background http.stop();  return "OK"; },
      "help": "shuts down this server",
      "logopt": 0,
      ),
);

# a logging closure
code log = sub (string str) {printf("%y: %s\n", now_us(), str);};

# our bind address
const Bind = 8888;

YamlRpcHandler yamlRpcHandler(new AbstractAuthenticator(), api);
our HttpServer http(log, log);
http.addListener(Bind);
http.setHandler("yamlrpc", "", MimeTypeYamlRpc, yamlRpcHandler);
http.setDefaultHandler("yamlrpc", yamlRpcHandler);
log("now listening on %s\n", Bind);
http.waitStop();
    @endcode
*/

//! the %YamlRpcHandler namespace holds all public definitions in the %YamlRpcHandler module
namespace YamlRpcHandler {
    //! YamlRpcHandler class definition; to be registered as a handler in the HttpServer class
/**
     */
class YamlRpcHandler : public AbstractHttpRequestHandler {

public:
        //! implementation of the handler
        const Version = "1.0";

        //! internal methods of the handler (introspection)
        const InternalMethods = (
            ("function": "help",
             "help"    : "shows a list of YAML-RPC methods registered with this handler",
             "text"    : "help",
             "logopt"  : 2
            ),
            ("function": "listMethods",
             "help"    : "lists YAML-RPC method names registered with this handler",
             "text"    : "system.listMethods",
             "logopt"  : 2
            ),
            ("function": "system_describe",
             "help"    : "returns a service description object, like the one specified by the JSON-RPC 1.1 spec (even though we use YAML instead of JSON)",
             "text"    : "system.describe",
             "logopt"  : 2
            ),
            );

        //! @cond nodoc
        private :
            list methods = ();
            hash mi;
            int loglevel;

            // if True then verbose exception info will be logged
            bool debug;

            // prefix to add to derived methods with GET requests if no "." characters are in the path
            __7_ string get_prefix;

            // a closure/call reference to get the log message and/or process arguments in incoming requests
            __7_ code getLogMessage;

            // a closure/call reference for logging (when set this is used instead of the HTTP server's logfunc for logging)
            __7_ code clog;

public:
        //! @endcond

        //! creates the handler with the given method list
/** @param v_auth an authentication object (use new AbstractAuthenticator() for no authentication)
            @param v_methods a list of hashes with the following keys:
            - \c name: a regular expression to use for matching the method name
            - \c function: a string (giving a function name to call), a call reference, or a closure to call with the deserialized arguments to the method; the return value will be serialized to YAML-RPC and sent back to the caller
            - \c help: help text for the method
            - \c text: the human-readable name of the method
            - \c logopt: (optional - by convention) log options which can be used by a custom logger (see the getLogMessage parameter)
            @param v_get_log_msg an optional closure or call reference to be called when an incoming request is received; if this is set then it will be called with the following arguments: a context hash (see HttpServer::AbstractHttpRequestHandler::handleRequest() for a description of the context hash), the method definition as passed in the methods argument to this constructor, and a reference to the arguments in this call
            @param v_dbg this parameter is set to @ref Qore::True "True", then additional information will be logged when errors occur
            @param v_get_prefix prefix to add to derived methods with GET requests if no "." characters are in the path
            @param v_log an optional closure or call reference to be called when logging

            @throw YAML-RPC-CONSTRUCTOR-ERROR missing \c "name", \c "function", or \"text" key in method hash, \c "function" key not assigned to a callable value
         */
        constructor(HttpServer::AbstractAuthenticator v_auth, list v_methods, __7_ code v_get_log_msg, bool v_dbg = False, __7_ string v_get_prefix, __7_ code v_log) ;


        //! adds a method to the handler dynamically
/** @param name a regular expression to use for matching the method name
            @param func a string (giving a function name to call), a call reference, or a closure to call with the deserialized arguments to the method; the return value will be serialized to YAML-RPC and sent back to the caller
            @param text the human-readable name of the method
            @param help help text for the method
            @param logopt log options which can be used by a custom logger (see the getLogMessage parameter)
            @param cmark any value that will be added in the method hash under the \c "cmark" key
          */
        addMethod(string name, code func, string text, string help, any logopt, any cmark);


        //! turns on or off debugging; when debugging is enabled more verbose error messages are reported
        setDebug(bool dbg = True);


        //! returns the current status of the debug flag
        bool getDebug();


        //! serializes a reponse in YAML-RPC format given the arguments
          static string makeResponse(any response, int flags = YAML::None);

        //! serializes an error reponse in YAML-RPC format given the arguments
          static string makeErrorResponse(int code, string mess, any err, int flags = YAML::None);

        //! @cond nodoc
        // method called by HttpServer to handle a request
        // don't reimplement this method; fix/enhance it in the module
        final hash handleRequest(hash cx, hash hdr, __7_ data body);


        // don't reimplement this method; fix/enhance it in the module
        final private addMethodInternal(hash h);


        private hash help();


        private hash system_describe();


        private list listMethods();


        private log(hash cx, string str);


        // don't reimplement this method; fix/enhance it in the module
        final private hash callMethod(hash cx, any params);

        //! @endcond
    };
};
